/***************************************************************************
 * Copyright (c) Fu Yin                                                     *
 *                                                                          *
 * Distributed under the terms of the BSD 3-Clause License.                 *
 *                                                                          *
 * The full license is in the file LICENSE, distributed with this software. *
 ****************************************************************************/
#include <iostream>
#include <string>
#include <vector>
#include <kfr/base.hpp>
#include <kfr/dsp.hpp>

#ifndef UTILS_HPP
#define UTILS_HPP

#ifdef _OPENMP
#include <omp.h>
#endif

// #include <ctime>
#include <chrono>
#include <exception>

#include <xtensor/xarray.hpp>
#include <xtensor/xtensor.hpp>
#include <xtensor/xbuilder.hpp>
#include <xtensor/xview.hpp>
#include <xtensor/xmath.hpp>
#include <xtensor/xcomplex.hpp>
#include <xtensor/xmanipulation.hpp>
#include <xtensor/xadapt.hpp>
#include <xtensor/xsort.hpp>
#include <xtensor/xoperation.hpp>
#include <xtensor/xindex_view.hpp>
#include <xsimd/xsimd.hpp>
#include <xtensor-blas/xlinalg.hpp>
#include <xtensor-python/pyarray.hpp>
#include <xtensor-python/pytensor.hpp>



#endif